declare name        "Ambisonic Panner";
declare version     "1.0";
declare author      "Pierre Lecomte";
declare license     "CC-BY-NC-SA-4.0";
declare copyright   "(c) Pierre Lecomte";
declare options "[osc:on]";

//###Panner###
// This tool allows to pan $S$ sources on $N$ loudspeakers using Ambisonic equivalent panning laws[^lecomte].
// The $i$-th source, with $i \in \\{1, \cdots, S\\}$ carries a signal denoted $s_i(z)$ in the discrete domain. 
// Encoded as a point source, its position is $(r_i,\theta_i,\phi_i)$ from origin and it emits a spherical wave.
// For the $n$-th loudspeaker with $n \in \\{1, \cdots, N\\}$ and coordinates $(r_n, \theta_n, \varphi_n)$, the driving signal $s_n$ is given by:
//
// $$\begin{equation}
// s_n = \sum\limits_{i=0}^{S} \sum_{l=0}^{L} s_i(z) w_{\text{max-}r_E, l}(L) \frac{F_l(r_i, z)}{F_l(r_n, z)}  (2 l + 1) P_l(\cos(\gamma_i(n))), 
// \label{eq:panning_law}
// \end{equation}$$
//
// where $P_l$ is the $l$-th Legendre polynomial, $w_{\text{max-}r_E}$ are the [max-$r_E$ weights]({% link docs/ylm.md %}#wrelmax-l1-l),
// $\frac{F_l(r_i, z)}{F_l(r_n, z)}$ are the The Near Field [(NF) filters]({% link docs/radial.md %}#stabilization-of-nf-filters-with-nfc-filters)),
// and $\gamma_i(n) = \cos(\phi_i) \cos(\phi_n) \cos(\theta_i - \theta_n) + \sin(\phi_i) \sin(\phi_n)$ is the angle between the $i$-th source and the $n$-th loudspeaker.
//
// The NF filters can be activated or not at compilation time with parameter `nfon`. 
// If activated (`nfon=1`), the gain attenuation and propagation delay between loudspeakers are as well equalized. 
// If not activated (`nfon=0`), $\frac{F_l(r_i, z)}{F_l(r_n, z)} = 1$ in Eq. \eqref{eq:panning_law}, that is to say that no near field effect is included in the process.
//
// ## Compilation parameters
// - `S`: number of source, $S > 0$
// - `L`: maximal Spherical Harmonics degree, $L > 0$,
// - `N`: number of loudspeaker, $N > 0$,
// - `nfon`: activate or not NF filters: `nfon=0` for no NF, `nfon=1` for NF.
// - `speaker(n) = (x, y, z)` $n$-th loudspeaker Cartesian coordinates in meters. One loudspeaker per line.
//
// ## Inputs / Outputs
// - Inputs: $S$
// - Outputs: $N$
//
// ## User Interface
//
// |             Element         |     OSC        | Min value | Max value | Step |
// |:---------------------------:|:--------------:|:---------:|:---------:|:----:|
// |          Gain (dB)          |   `gain_i`     |    -20    |    20     |  0.1 |
// |        Radius $r$ (m)       |  `radius_i`    |    0.75   |    50     | 0.01 |
// | Azimuth $\theta$ ($^\circ$) | `azimuth_i`    |    -180   |    180    | 0.01 |
// | Elevation $\phi$ ($^\circ$) |  `elevation_i` |    -90    |    90     | 0.01 |
//
//
// [^lecomte]: P. Lecomte, P.-A. Gauthier, C. Langrenne, A. Berry, et A. Garcia, « A Fifty-Node Lebedev Grid and Its Applications to Ambisonics », Journal of the Audio Engineering Society, vol. 64, nᵒ 11, p. 868‑881, 2016.


import("stdfaust.lib");
import("ylm.lib");
import("grids.lib");
import("radial.lib");

// COMPILATION PARAMETERS
S   =   2;  // Source number
L   =   3;  // Maximum required degree $L$
N   =   22; // Loudspeaker number
nfon = 1;

// Loudspeakers Cartesian coordinates $(x, y, z)$ in meters.
speaker(0) = (1.6, 2, 0);
speaker(1) = (1.6, 1, 0);
speaker(2) = (1.6, 0, 0);
speaker(3) = (1.6, -1, 0);
speaker(4) = (1.6, -2, 0);
speaker(5) = (0, 2, 0);
speaker(6) = (0, -2, 0);
speaker(7) = (-1.7, 1.7, 0);
speaker(8) = (-1.7, 0, 0);
speaker(9) = (-1.7, -1.7, 0);
speaker(10) = (1.6, 2, 1.24);
speaker(11) = (1.6, 0, 1.24);
speaker(12) = (1.6, -2, 1.24);
speaker(13) = (0, 2, 1.24);
speaker(14) = (0, 0, 1.24);
speaker(15) = (0, -2, 1.24);
speaker(16) = (-1.7, 1.7, 1.24);
speaker(17) = (-1.7, 0, 1.24);
speaker(18) = (-1.7, -1.7, 1.24);
speaker(19) = (1.6, 2, -1);
speaker(20) = (1.6, 0, -1);
speaker(21) = (1.6, -2, -1);


// DO NOT EDIT BELOW HERE

rs = par(i, N, ba.take(1, cart2spher(ba.take(1, speaker(i)), ba.take(2, speaker(i)), ba.take(3, speaker(i)))));
rmax = sup(rs);

// User interface
g(s)	=	hslider("[%s+1][unit:dB][osc:/gain_%s -20 20][style:knob]Gain %2s",0,-20,20,0.1): ba.db2linear : si.smoo; // gain
r(s)	=	hslider("[%s+2][unit:m][osc:/radius_%s 0.5 50][style:knob]Radius %2s", 1, 0.75, 50, 0.01) : si.smoo;// radius
t(s)	=	hslider("[%s+3][unit:°][osc:/azimuth_%s -180 180][style:knob]Azimuth %2s", 0, -180, 180, 0.1)*ma.PI/180 : si.smoo; // azimuth
p(s)	=	hslider("[%s+4][unit:°][osc:/elevation_%s -90 90][style:knob]Elevation %2s", 0, -90, 90, 0.1)*ma.PI/180 : si.smoo; // zenith

gain(s, i) = g(s) * case{
                (0) => sum(l, L + 1, nf(l, r(s), r1) * (2 * l + 1) * wre(L, L, l) * legendre(l, costheta(t1, p1, t(s), p(s) ))) / rmax : delay((rmax - r1)); // normalize by rmax
                (1) => _ * sum(l, L + 1, (2 * l + 1) * wre(L, L, l) * legendre(l, costheta(t1, p1, t(s), p(s) )));
                }(nfon==0)  with 
                        {
                        r1 = ba.take(1, cart2spher(ba.take(1, speaker(i)), ba.take(2, speaker(i)), ba.take(3, speaker(i))));
                        t1 = ba.take(2, cart2spher(ba.take(1, speaker(i)), ba.take(2, speaker(i)), ba.take(3, speaker(i))));
                        p1 = ba.take(3, cart2spher(ba.take(1, speaker(i)), ba.take(2, speaker(i)), ba.take(3, speaker(i))));
                        };


process = par(s, S, hgroup("Source %2s", _<: par(i, N, gain(s, i)))) :> si.bus(N);

